from setup import __version__

from braga.entity import Entity
from braga.component import Component
from braga.system import System
from braga.assemblage import Assemblage
from braga.aspect import Aspect
from braga.world import World



import abc
import threading

from braga.aspect import Aspect


class System(object):
    """Handler for modifying and updating Entities with a particular pattern of Components."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, world, aspect=None):
        self.thread = None
        self.world = world
        self.aspect = aspect if aspect else Aspect()

    def start(self):
        self.thread = threading.Thread(name=type(self).__name__, target=self.update)
        self.thread.start()

    def __contains__(self, entity):
        return entity in self.aspect

    @property
    def entities(self):
        """Set of Entities in the World with this Aspect"""
        return self.world.entities_with_aspect(self.aspect)

    @abc.abstractmethod
    def update(self):
        """Updates the entities in this system"""



import uuid


class Entity(object):
    """Represents a unique object."""

    def __init__(self):
        self._uuid = uuid.uuid4()
        self.components = set()

    @property
    def uuid(self):
        """Property to access the Entity's uuid, exists to prevent uuid modification."""
        return self._uuid

    def get_component(self, component_type):
        """ Retrives the component of a particular type belonging to this Entity.
            Raises an AttributeError if the Entity does not have a component of that class.
        """

        components = [comp for comp in self.components if isinstance(comp, component_type)]
        if not components:
            raise AttributeError
        else:
            return components[0]

    def has_component(self, component_type):
        """Determines if the Entity has a component of the given type."""
        return any(isinstance(comp, component_type) for comp in self.components)

    def __repr__(self):
        return "{0}({1}) - {2}".format(type(self).__name__, self._uuid, self.components)

    def __getattr__(self, name):
        """ Checks components for an attribute with a given name.
            Only raise an AttributeError if none of the Entity's components
            have an attribute with that name.
        """

        for component in self.components:
            try:
                attr = getattr(component, name)
            except AttributeError:
                pass
            else:
                return attr
        raise AttributeError



from collections import defaultdict

from braga import Entity


class Assemblage(object):
    """Factory for making Entities with particular combinations of Components."""

    def __init__(self, components=None):
        """ `components` can either be a list of Component classes
            or a dictionary, where the keys are
            Component classes and the values are dictionaries, where
            the key-value pairs are kwargs for the __init__ methods of the Component class
        """
        self.component_types = defaultdict(dict)
        if components and isinstance(components, dict):
            self.component_types.update(components)
        elif components and isinstance(components, list):
            for component in components:
                self.component_types[component]

    def add_component(self, component_type, **kwargs):
        self.component_types[component_type].update(**kwargs)

    def make(self, **kwargs):
        """ Makes an Entity with the Assemblage's combination of Components.

            Can initialize the Entity's components with particular values.
            If the Assemblage has been created with initial values for the parameters of its Components,
                those initial values are applied.
            If this method is called with kwargs, it will pass that initial value to the appropriate Component.
            Calling `make` with kwargs will override the initial values set on the Assemblage.

            Returns an Entity.
        """

        entity = Entity()

        for component_type, init_kwargs in self.component_types.iteritems():
            instance_kwargs = init_kwargs
            instance_kwargs.update({k:v for k,v in kwargs.iteritems() if k in component_type.INITIAL_PROPERTIES})

            component = component_type(**instance_kwargs)
            entity.components.add(component)

        return entity



class Aspect(object):
    """ Defines a pattern of Component types. Can be used to check if the set of components
        belonging to an Entity is consistent with the pattern.
    """

    def __init__(self, all_of=None, exclude=None, some_of=None):
        """ Defines the pattern of Components for this Aspect.
            :param all_of: Components that must be present
            :type all_of: set.
            :param exclude: Components that must not be present
            :type exclude: set.
            :param some_of: Components that may be present
            :type some_of: set.
        """
        self.all_of = all_of if all_of else set()
        self.exclude = exclude if exclude else set()
        self.some_of = some_of if some_of else set()

    def is_interested_in(self, entity):
        """ Determines if an Entity has Components consistent with the Aspect's pattern.
        :param entity: the Entity to evaluate
        :type entity: Entity.
        :returns: bool -- whether the Entity matches the Aspect
        """
        return (self._is_interested_in_all_of(entity) and
                self._is_interested_in_exclude(entity) and
                self._is_interested_in_some_of(entity))

    def __contains__(self, entity):
        """Syntactic sugar for `is_interested_in`."""
        return self.is_interested_in(entity)

    def _is_interested_in_all_of(self, entity):
        components = set([type(component) for component in entity.components])
        return components.issuperset(self.all_of)

    def _is_interested_in_exclude(self, entity):
        components = set([type(component) for component in entity.components])
        return not components & self.exclude

    def _is_interested_in_some_of(self, entity):
        components = set([type(component) for component in entity.components])
        return components & self.some_of or not self.some_of

    def select_entities(self, entities):
        return set([entity for entity in entities if self.is_interested_in(entity)])



from setuptools import setup

__version__ = '0.5.5'

if __name__ == '__main__':
    setup(name='braga',
          version=__version__,
          description='Toy Entity-Component system',
          url='http://github.com/astrosilverio/braga',
          author='astrosilverio',
          author_email='astrosilverio@gmail.com',
          license='MIT',
          packages=['braga'],
          zip_safe=False)



from braga.entity import Entity


class World(object):
    """Collects all the Entities for a particular program."""

    def __init__(self):
        self.entities = set()

    def entities_with_aspect(self, aspect):
        """Returns a set of Entities in the World with a particular Aspect."""
        return set([entity for entity in self.entities if entity in aspect])

    def make_entity(self, assemblage=None, **kwargs):
        """ Creates an Entity for this World.
            :param assemblage: Assemblage the Entity should be built with
            :type assemblage: Assemblage.

            Can take kwargs that will be passed to the assemblage's `make` method.

            Returns an Entity.
        """

        if not assemblage:
            new_entity = Entity()
        else:
            new_entity = assemblage.make(**kwargs)

        self.entities.add(new_entity)
        return new_entity
